# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import numpy as nmpy
from cell_tracking_BC.type.frame import frame_t
from cell_tracking_BC.type.segmentation import segmentation_t


array_t = nmpy.ndarray


shape = (100, 100)

nucleus_map = nmpy.zeros(shape, dtype=nmpy.bool_)
nucleus_map[50:58, 60:72] = True

cytoplasm_map = nmpy.zeros(shape, dtype=nmpy.bool_)
cytoplasm_map[45:65, 54:80] = True
cytoplasm_map[nucleus_map] = False

frame = frame_t(nucleus_map + cytoplasm_map)
segmentation = segmentation_t.NewFromCompartments(
    nucleus=nucleus_map, cytoplasm=cytoplasm_map
)
frame.AddCellsFromSegmentation(segmentation)
print(frame)
